<?php

namespace PowerFlux;


enum ERRORCODES: int
{
    case HTTP401 = 401;
    case HTTP400 = 400;
    case HTTP403 = 403;
    case HTTP404 = 404;
    case HTTP429 = 429;
    case HTTP503 = 503;

    case APP_NOREFRESH = 801;
    case APP_INVALID_DATEFROM = 802;
    case APP_INVALID_DATETO = 803;
    case APP_INVALID_TIMEAGGR = 804;
    case APP_INVALID_METERID = 805;
    case APP_NOSETTINGS = 806;
    case APP_INVALID_RESPONSE = 807;

    case APP_INVALID_INFLUXDB_URL = 808;
    case APP_INVALID_INFLUXDB_TOKEN = 809;
    case APP_INVALID_INFLUXDB_BUCKET = 810;
    case APP_INVALID_INFLUXDB_ORG = 811;
    case APP_NOAPPFILE = 812;

    case APP_INVALID_QUERY_DATEFROM = 813;
    case APP_INVALID_QUERY_DATETO = 814;

    case INFLUX_GETLATEST = 851;
    case INFLUX_WRITE = 852;

    case UNKNOWN = 999;
}

enum ERRORLEVELS: int
{
    case DEBUG = 8;
    case INFO = 7;
    case NOTICE = 6;
    case WARNING = 5;
    case ERROR = 4;

    public static function getString(int $int)
    { //enum to string
        switch (self::from($int)) {

            case ERRORLEVELS::DEBUG:
                return "DEBUG";
                break;

            case ERRORLEVELS::INFO:
                return "INFO";
                break;

            case ERRORLEVELS::NOTICE:
                return "NOTICE";
                break;

            case ERRORLEVELS::WARNING:
                return "WARNING";
                break;

            case ERRORLEVELS::ERROR:
                return "ERROR";
                break;
        }
    }
}

class ErrorOutput
{
    private static $_instance;
    private static $loglevel;

    public static function setLogLevel(string $level)
    { //String to enum
        switch ($level) {

            case "DEBUG":
                self::$loglevel = ERRORLEVELS::DEBUG;
                break;

            case "INFO":
                self::$loglevel = ERRORLEVELS::INFO;
                break;

            case "NOTICE":
                self::$loglevel = ERRORLEVELS::NOTICE;
                break;

            case "WARNING":
                self::$loglevel = ERRORLEVELS::WARNING;
                break;

            case "ERROR":
                self::$loglevel = ERRORLEVELS::ERROR;
                break;

            default:
                self::$loglevel = ERRORLEVELS::ERROR;
        }
    }

    public static function getLogLevel()
    {
        return self::$loglevel;
    }

    #region LOG methods (ERROR levels)
    public static function DEBUG($str)
    {
        self::doLog(level: ERRORLEVELS::DEBUG, input: $str);
    }

    public static function INFO($str)
    {
        self::doLog(level: ERRORLEVELS::INFO, input: $str);
    }

    public static function NOTICE($str)
    {
        self::doLog(level: ERRORLEVELS::NOTICE, input: $str);
    }

    public static function WARNING($str)
    {
        self::doLog(level: ERRORLEVELS::WARNING, input: $str);
    }

    public static function ERROR($str)
    {
        self::doLog(level: ERRORLEVELS::ERROR, input: $str);
    }
    #endregion

    private static function doTrace(bool $echo = false): string|bool
    {
        if ($echo) {
            $bt = debug_backtrace(!DEBUG_BACKTRACE_PROVIDE_OBJECT | DEBUG_BACKTRACE_IGNORE_ARGS);
            $i = 0;
            foreach ($bt as $stack) {
                if ($stack["function"] !== "handleError" && $stack["function"] !== "doTrace") { //Skip
                    $i++;
                    echo "TRACE (" . $i . "): Called " . $stack["function"] . " (" . $stack["class"] . ") from line " . $stack["line"] . " in " . $stack["file"] . PHP_EOL;
                }
            }
            echo PHP_EOL;
            return (string)$bt[2]["function"] ?? false;
        } else {
            $bt = debug_backtrace(!DEBUG_BACKTRACE_PROVIDE_OBJECT | DEBUG_BACKTRACE_IGNORE_ARGS, 4);
            return (string)$bt[3]["function"] ?? false;
        }
    }




    /**
     * Appends log entry to file. (if logging is enabled)
     */
    public static function doLog(ERRORLEVELS $level, $input)
    {
        $func = ErrorOutput::doTrace();
        if (Configuration::getToggleLog()) {
            if ($level->value <= self::$loglevel->value) {
                $str = Common::getTimeDate(format: "d-m-Y H:i:s", tz: "Europe/Copenhagen") . "[" . ERRORLEVELS::getString($level->value) . "] (" . $func . ") " . $input;
                file_put_contents(Configuration::getFilePath("log"), $str . PHP_EOL, FILE_APPEND);
            }
        }
    }

    public static function handleError(ERRORCODES $error, $exc = "")
    {
        $func = ErrorOutput::doTrace();

        switch ($error) {

            case ERRORCODES::HTTP401:
                if ($func === "fetchToken") {
                    $infostr = "Couldn't authenticate with refresh token, please check that the refresh token is in fact valid and active.";
                    Configuration::setAppEntry("ts_api_readings_lastfail", Common::getTime());
                    ErrorOutput::ERROR($infostr);
                } else {
                    $infostr = "Unhandled error occured in " . $func . " - 401 Unauthorized";
                    ErrorOutput::ERROR($infostr);
                }
                break;

            case ERRORCODES::HTTP400:
                if ($func === "fetchMeterReadings") {
                    $infostr = "You don't have access to the meter IDs set in settings.json.";
                    echo $infostr . PHP_EOL;
                    ErrorOutput::WARNING($infostr);
                } else {
                    $infostr = "Unhandled error occured in " . $func . " - 400 Bad Request";
                    ErrorOutput::WARNING($infostr);
                }
                break;

            case ERRORCODES::HTTP403:
                $infostr = "Wrong type of token was used. This should never happen, contact the developer.";
                ErrorOutput::INFO($infostr);
                break;

            case ERRORCODES::HTTP404:
                $infostr = "Error occured in " . $func . " - 404 Not found"; //Only log - Most likely cause is malformed request url(fetchmeterreadings)
                //echo $infostr;
                ErrorOutput::WARNING($infostr);
                exit($infostr);
                break;

            case ERRORCODES::HTTP429:
                $infostr = "Too many requests - This should normally not happen since the code delays requests to Eloverblik as set in Configuration.php, if it continues contact the developer.";
                Configuration::setAppEntry("ts_api_readings_lastfail", Common::getTime());
                //echo $infostr;
                ErrorOutput::WARNING($infostr);
                break;

            case ERRORCODES::HTTP503:
                $infostr = "DataHub is unavailable.";
                Configuration::setAppEntry("ts_api_readings_lastfail", Common::getTime());
                ErrorOutput::WARNING($infostr);
                break;

            case ERRORCODES::APP_NOREFRESH:
                $infostr = "No refresh token was provided. Please check settings.json.";
                ErrorOutput::ERROR($infostr);
                exit($infostr);
                break;

            case ERRORCODES::APP_INVALID_DATEFROM:
                $infostr = "The field date_from is empty, missing or in a invalid format. Please check settings.json.";
                ErrorOutput::ERROR($infostr);
                exit($infostr);
                break;

            case ERRORCODES::APP_INVALID_DATETO:
                $infostr = "The field date_to is empty, missing or in a invalid format. Please check settings.json.";
                ErrorOutput::ERROR($infostr);
                exit($infostr);
                break;

            case ERRORCODES::APP_INVALID_TIMEAGGR:
                $infostr = "The field time_aggregation is empty, missing or has a invalid value. Please check settings.json.";
                ErrorOutput::ERROR($infostr);
                echo $infostr . PHP_EOL;
                break;

            case ERRORCODES::APP_INVALID_METERID:
                $infostr = "The field meter_id is empty, missing or has an invalid value. Please check settings.json.";
                ErrorOutput::ERROR($infostr);
                exit($infostr);
                break;

            case ERRORCODES::APP_INVALID_RESPONSE:
                $infostr = "Received invalid or empty response from Eloverblik.";
                Configuration::setAppEntry("ts_api_readings_lastfail", Common::getTime());
                ErrorOutput::NOTICE($infostr);
                break;

            case ERRORCODES::APP_NOSETTINGS:
                $infostr = "No settings file found, creating new.";
                echo $infostr . PHP_EOL;
                ErrorOutput::INFO($infostr);
                break;

            case ERRORCODES::APP_NOAPPFILE:
                $infostr = "No app file found, creating new.";
                echo $infostr . PHP_EOL;
                ErrorOutput::INFO($infostr);
                break;

            case ERRORCODES::APP_INVALID_QUERY_DATEFROM:
                $infostr = "The field query_date_from in app.json is empty or invalid - Should not happen.";
                ErrorOutput::ERROR($infostr);
                exit($infostr);
                break;

            case ERRORCODES::APP_INVALID_QUERY_DATETO:
                $infostr = "The field query_date_to in app.json is empty or invalid - Should not happen.";
                ErrorOutput::ERROR($infostr);
                exit($infostr);
                break;

            case ERRORCODES::APP_INVALID_INFLUXDB_URL:
                $infostr = "The field influxdb_url is empty, missing or has an invalid value. Please check settings.json.";
                ErrorOutput::ERROR($infostr);
                exit($infostr);
                break;

            case ERRORCODES::APP_INVALID_INFLUXDB_TOKEN:
                $infostr = "The field influxdb_token is empty, missing or has an invalid value. Please check settings.json.";
                ErrorOutput::ERROR($infostr);
                exit($infostr);
                break;

            case ERRORCODES::APP_INVALID_INFLUXDB_BUCKET:
                $infostr = "The field influxdb_bucket is empty, missing or has an invalid value. Please check settings.json.";
                ErrorOutput::ERROR($infostr);
                exit($infostr);
                break;

            case ERRORCODES::APP_INVALID_INFLUXDB_ORG:
                $infostr = "The field influxdb_org is empty, missing or has an invalid value. Please check settings.json.";
                ErrorOutput::ERROR($infostr);
                exit($infostr);
                break;

            case ERRORCODES::INFLUX_GETLATEST:
                $infostr = "Something went wrong while trying to get the latest entry from the database.";
                ErrorOutput::ERROR($infostr);
                exit($infostr);
                break;

            case ERRORCODES::INFLUX_WRITE:
                $infostr = "Something went wrong while trying to write data to InfluxDB.";
                ErrorOutput::ERROR($infostr);
                exit($infostr);
                break;

            case ERRORCODES::UNKNOWN:
                $infostr = "Unknown error occured, in: " . $func . " -> " . $exc;
                ErrorOutput::ERROR($infostr);
                break;

            default: //Should never hit this
                ErrorOutput::ERROR("Something went wrong in error handling, please report.");
        }
    }


    public function __construct()
    {
        self::$_instance = $this;
    }

    public static function getInstance()
    {
        if (self::$_instance == null) {
            self::$_instance = new ErrorOutput();
        }
        return self::$_instance;
    }
}
