<?php
namespace PowerFlux;

use DateTimeImmutable;
use DateTimeZone;
use InfluxDB2\Client;
use InfluxDB2\Model\WritePrecision;
use InfluxDB2\WriteType;

class Influxdb {

    public static function getClient()
    {
        try {
            $url = Configuration::getSettingsEntry("influxdb_url") ?? ErrorOutput::handleError(ERRORCODES::APP_INVALID_INFLUXDB_URL ?? ERRORCODES::UNKNOWN);
            $token = Configuration::getSettingsEntry("influxdb_token") ?? ErrorOutput::handleError(ERRORCODES::APP_INVALID_INFLUXDB_TOKEN ?? ERRORCODES::UNKNOWN);
            $bucket = Configuration::getSettingsEntry("influxdb_bucket") ?? ErrorOutput::handleError(ERRORCODES::APP_INVALID_INFLUXDB_BUCKET ?? ERRORCODES::UNKNOWN);
            $org = Configuration::getSettingsEntry("influxdb_org") ?? ErrorOutput::handleError(ERRORCODES::APP_INVALID_INFLUXDB_ORG ?? ERRORCODES::UNKNOWN);

            $client = new Client([
            "url" => $url,
            "token" => $token,
            "bucket" => $bucket,
            "org" => $org,
            "precision" => WritePrecision::S
            ]);
            if ($client) {
                return $client;
            }

        } catch (\Exception $e) {
            ErrorOutput::handleError(ERRORCODES::UNKNOWN);
        }
    }

public static function isValidURL($str): string|false { //Todo: move to common //Todo: Checks that this is somewhat valid
    if (strlen($str) >= 10) {
        return $str;
    }
    return false;
}

public static function isValidToken($str): string|false {
    if (strlen($str) >= 15) {
        return $str;
    }
    return false;
}

public static function isValidOrg($str): string|false {
    if (strlen($str) >= 1) {
        return $str;
    }
    return false;
}

public static function isValidBucket($str): string|false {
    if (strlen($str) >= 1) {
        return $str;
    }
    return false;
    //Todo: Check if bucket exists in InfluxDB
}

    public static function getLatestReading($meter = "", $type = "readings"): ?DateTimeImmutable
    {
        if ($client = Influxdb::getClient()) {
            $queryApi = $client->createQueryApi();
            $bucket = Configuration::getSettingsEntry("influxdb_bucket");

            switch ($type) {
                
                case "readings":
                    $str = <<<HEREDOC
                    from(bucket: "$bucket" )
                    |> range(start: -50y)
                    |> filter(fn: (r) => r["_measurement"] == "readings")
                    |> filter(fn: (r) => r["_field"] == "value")
                    |> filter(fn: (r) => r["meter"] == "$meter")
                    |> last()
                    HEREDOC;
                break;

                case "prices":
                    $str = <<<HEREDOC
                    from(bucket: "$bucket" )
                    |> range(start: -50y)
                    |> filter(fn: (r) => r["_measurement"] == "prices")
                    |> filter(fn: (r) => r["_field"] == "value")
                    |> last()
                    HEREDOC;
                break;

            }
            
            $query = $queryApi->query($str);

            if ( isset($query[0]->records[0]->values["_time"]) ) {
                $lastreading_ts = $query[0]->records[0]->values["_time"];
                $tmp = true; //Todo: Check if var contains valid datetime

                if ($tmp) {
                    ErrorOutput::DEBUG("Last entry in database (type: " . $type . "): " . Common::getDateTime($lastreading_ts)->format(Common::DATEHUMAN));

                    $r = (new \DateTimeImmutable($lastreading_ts, new DateTimeZone("UTC")));
                }
            }
        }
        return $r ?? null;
    }

    public static function doAncientPeriods($periods, $meter) {
        
        if ( (count($periods)) !== 0) {
        foreach ($periods as $period) {
            ErrorOutput::DEBUG("---------------------------");

            if (!Configuration::doIgnore(type: "readings", year: Common::getTimeDate(str: $period["start"], format: "Y")) ) { //If year is not in ignore list
                ErrorOutput::DEBUG("FALSE from doIgnore() - START: " . $period["start"] . "   ->   " . $period["stop"]);

                $period_start = Common::getTimeDate(str: $period["start"]);
                $period_end = Common::getTimeDate(str: $period["stop"]);

                if ($r = Parser::doTimeSeries(Parser::getAncientMeterReadings(datefrom: $period_start, dateto: $period_end))) {
                    ErrorOutput::INFO("RESPONSE FROM getAncientMeterReadings() -> doWrite()");
                    if (Influxdb::doWrite($meter, $r) ) {
                        Configuration::addToIgnore(type: "readings", year: Common::getTimeDate(str: $period["start"], format: "Y"));
                    }
                }
            }
        }
    } else {
        ErrorOutput::DEBUG("No periods left to do.");
    }
    }


    /** Writes meter readings to InfluxDB database.
     * 
     */
    public static function doWrite($meter, $input): ?bool
    {
        $r = false;
        if ($client = Influxdb::getClient()) {
            $data = array();
            $dataout = array();
            $writeApi = $client->createWriteApi(["writeType" => WriteType::BATCHING, 'batchSize' => 800]);

            $data["name"] = "readings";
            $data["tags"] = [ "meter" => strval($meter)  ];
            ErrorOutput::DEBUG("WRITE - FOR METER: " . json_encode($data["tags"]["meter"],JSON_PRETTY_PRINT));

            foreach ($input["data"] as $hour) {
                $timestamp = $hour["date"];
                $value["value"] = floatval($hour["kwh"]);
                $data["time"] = strtotime($timestamp);

                $data["fields"] = (array)$value;
                array_push($dataout, $data);
                $writeApi->write($data);
            }

            //todo: tmp, remove->get count from db or log from write instance
            if ((count($input["data"]) == count($dataout)) && (count($dataout) >= 2)  ) {
                $r = true;
            }
            ErrorOutput::INFO("WRITE - Entries: " . count($dataout));
            ErrorOutput::DEBUG("WRITE - First entry: " . json_encode($dataout[array_key_first($dataout)] ?? null,JSON_PRETTY_PRINT));
            ErrorOutput::DEBUG("WRITE - Last entry: " . json_encode($dataout[array_key_last($dataout)] ?? null, JSON_PRETTY_PRINT));
            $writeApi->close();
        }
        return $r ?? ErrorOutput::handleError(ERRORCODES::INFLUX_WRITE);
    }



    /** Writes electricity prices to InfluxDB database.
     * 
     */
    public static function doWriteForPrices($input): ?bool {
        if ($client = Influxdb::getClient()) {
            $data = array();
            $dataout = array();
            $writeApi = $client->createWriteApi(["writeType" => WriteType::BATCHING, 'batchSize' => 800]);
            $data["name"] = "prices";

            foreach ($input["data"] as $hour) {
                $timestamp = $hour["date"];
                $value["value"] = floatval($hour["priceDKK"]);
                $data["tags"] = [ "pricearea" => strval($hour["priceArea"]) ];
                $data["time"] = strtotime($timestamp);

                $data["fields"] = (array)$value;
                array_push($dataout, $data);
                $writeApi->write($data);
            }


            //todo: see above (doWrite)
            if (count($input["data"]) == count($dataout)) {
                $r = true;
            }
            ErrorOutput::INFO("WRITE - Entries: " . count($dataout));
            ErrorOutput::DEBUG("WRITE - First entry: " . json_encode($dataout[array_key_first($dataout)] ?? null,JSON_PRETTY_PRINT));
            ErrorOutput::DEBUG("WRITE - Last entry: " . json_encode($dataout[array_key_last($dataout)] ?? null, JSON_PRETTY_PRINT));
            $writeApi->close();
        }
        return $r ?? ErrorOutput::handleError(ERRORCODES::INFLUX_WRITE);
    }

}

