<?php
namespace PowerFlux;

use DateTimeImmutable;
use DateTimeZone;
use PowerFlux\Files\File;
use GuzzleHttp\Client;
use GuzzleHttp\Psr7;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ServerException;

class Parser {

    private static $_instance;

    public function __construct()
    {
    }


    /**
     * Tries to get a access token from Eloverblik.
     */
    public static function fetchToken(): string|bool
    {
        $apiurl = Configuration::getApiUrl();

        if (!$refreshtoken = Configuration::getRefreshToken()) {
            ErrorOutput::handleError(ERRORCODES::APP_NOREFRESH ?? ERRORCODES::UNKNOWN);
        } else {

            if (Parser::doAttempt()) {
                try {
                    $client = new \GuzzleHttp\Client(['base_uri' => $apiurl]);
                    $response = $client->request("GET", "api/token", [
                        "headers" => [
                            "user-agent" => "guzzlehttp/1.0",
                            "accept"     => 'application/json',
                            "authorization"      => 'Bearer ' . $refreshtoken
                        ]
                    ]);

                    if ($response->getStatusCode() == "200") {

                        //Todo: Validation of token here
                        $tmp = false;
                        if ( strlen((string) $response->getBody()) > 10) {
                            $tmp = true;
                        }
                        //

                        if ($tmp) {
                            $body = (string) $response->getBody();
                            if ($body) {
                                $j = json_decode($body, true);
                                $token = $j["result"];
                                if ($token) {
                                    Configuration::setAccessToken($token);
                                    return $token;
                                }
                            }
                        } else { //Unable to validate token
                            ErrorOutput::handleError(ERRORCODES::APP_INVALID_RESPONSE ?? ERRORCODES::UNKNOWN );
                            return false;
                        }
                    }

                    return false; //
                } catch (ClientException $e) { //4xx responses
                    $exresp = $e->getResponse();
                    ErrorOutput::handleError(ERRORCODES::tryFrom($exresp->getStatusCode()) ?? ERRORCODES::UNKNOWN );
                    return false;
                }
                catch (ServerException $e) { //5xx responses
                    $exresp = $e->getResponse();
                    ErrorOutput::handleError(ERRORCODES::tryFrom($exresp->getStatusCode()) ?? ERRORCODES::UNKNOWN );
                    return false;
                }
            }
        }
        return false;
    }

    /**
     * Tries to get the available meter(s) from Eloverblik.
     */
    public static function fetchMeters()
    {
        $apiurl = Configuration::getApiUrl();
        if (($accesstoken = Configuration::getAccessToken()) && (Parser::doAttempt())) {

            try {
                $client = new \GuzzleHttp\Client(['base_uri' => $apiurl]);
                $response = $client->request("GET", "api/meteringpoints/meteringpoints", [
                    "headers" => [
                        "user-agent" => "guzzlehttp/1.0",
                        "accept"     => "application/json",
                        "authorization"      => "Bearer " . $accesstoken
                    ],
                    "query" => ["includeAll" => "false"]
                ]);
                ErrorOutput::DEBUG("CALL fetchMeters()");

                    //Not sure how data structure is when more than one metering point is present(probably values of 'childMeteringPoints'), so only return the first
                    if ($response->getStatusCode() == "200") {
                        ErrorOutput::DEBUG("RESPONSE: 200");
                        $resp = (string)$response->getBody();

                        if (!empty($resp)) { //Todo: clean up
                            $array = json_decode($resp, true);
                            if (array_key_exists("result", $array)) {

                                //Update date_from entry, if relevant
                                if (array_key_exists("consumerStartDate", $array["result"][0])) {
                                    $csd = $array["result"][0]["consumerStartDate"];
                                    //Allows date to be specificed in settings.json, as long as it is AFTER consumerStartDate
                                    if (Common::isDateAfter(Configuration::getSettingsEntry("date_from"), $csd)) {
                                        Configuration::setSettingsEntry("date_from", Common::getTimeDate($csd));
                                    }
                                }

                                if (array_key_exists("meteringPointId", $array["result"][0])) {
                                    return $array["result"][0]["meteringPointId"];
                                }
                            }
                        }
                        return false;
                    }

            } catch (ClientException $e) { //4xx responses
                $exresp = $e->getResponse();
                if ($exresp->getStatusCode() == "401") { //Unauthorized - Access token probably expired (valid for 24h)
                    ErrorOutput::DEBUG("Unauthorized - fetch new access token");
                    Parser::fetchToken(); //Request new access token
                } else {
                    ErrorOutput::handleError(ERRORCODES::tryFrom($exresp->getStatusCode()) ?? ERRORCODES::UNKNOWN);
                }
            }
            catch (ServerException $e) { //5xx responses
                $exresp = $e->getResponse();
                ErrorOutput::handleError(ERRORCODES::tryFrom($exresp->getStatusCode()) ?? ERRORCODES::UNKNOWN);
            }

            
        }
        return false;
    }

    /**
     * Checks if enough time has passed since last unsuccessful request to Eloverblik (90s)
     */
    private static function doAttempt(): bool {

        if ($ts = Configuration::getAppEntry("ts_api_readings_lastfail")) {
            $timediff = Common::getTimeDifference($ts);

            if ($timediff) {
                if ($timediff >= Configuration::REQUEST_DELAYPERIOD) {
                    return true;
                } else {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Returns untouched meter readings from data file.
     * Will attempt to fetch new data from Eloverblik(depending on time passed since last request)*
     */
    public static function getMeterReadings(string $datefrom = "", string $dateto = ""): ?string
    {
        $datats = Configuration::getAppEntry("ts_api_readings_resp") ?? 1;
        $timediff = round((Configuration::getUpdateIntervalforReadings()-(abs(Common::getTimeDiff( Common::getDateTimeFromTS($datats), as: "sec"))))/60,0);
        $timesince = abs(Common::getTimeDiff(Common::getDateTimeFromTS($datats), as: "sec"));
  
        ErrorOutput::DEBUG("DELAY PERIOD, NEXT FETCH IN: " . $timediff. " minutes");
        ErrorOutput::DEBUG("DELAY PERIOD, WAIT? -> LAST FETCH WAS " . round($timesince/60,0) . " minutes ago");
        ErrorOutput::DEBUG("GRACE PERIOD, DOATTEMPT: " . self::doAttempt() );

        if (($timesince > Configuration::getUpdateIntervalforReadings()) && (Parser::doAttempt())) { //Data is old & last attempt was more than X seconds ago, fetch new
            ErrorOutput::INFO("ACTION -> fetchMeterReadings : LAST FETCH WAS: " . round($timesince/60,0) . " minutes ago");
            $data = self::fetchMeterReadings(datefrom: $datefrom, dateto: $dateto);

            if ($data) {
                File::saveFile("dataresp", $data); //Todo, rework to fit new model
                Configuration::setAppEntry("ts_api_readings_resp", Common::getTime());
            } else {
                Configuration::setAppEntry("ts_api_readings_lastfail", Common::getTime()); //Store timestamp of last *unsuccessful* attempt to fetch data
            }
        } else {
            ErrorOutput::DEBUG("STOP: NOT RUNNING - WAITING " . $timediff. " minutes");
        }
        return $data ?? null;
    }

    // Like getMeterReadings() but skips getUpdateIntervalforReadings()
    public static function getAncientMeterReadings(string $datefrom = "", string $dateto = ""): ?string {
        if (Parser::doAttempt()) {
            $data = self::fetchMeterReadings(datefrom: $datefrom, dateto: $dateto);

                if ($data) {
                    File::saveFile("dataresp", $data);
                    Configuration::setAppEntry("ts_api_readings_resp", Common::getTime());
                    Configuration::setAppEntry("query_date_from", Common::getTimeDate(str: $datefrom));
                    Configuration::setAppEntry("query_date_to", Common::getTimeDate(str: $dateto));
                } else {
                    Configuration::setAppEntry("ts_api_readings_lastfail", Common::getTime()); //Store timestamp of last *unsuccessful* attempt to fetch data
                }
        }
        return $data ?? null;
    }

    /**
     * Returns array of meter readings (kWh per hour). Restructure of $input (expects full gettimeseries Eloverblik response)
     */
    public static function doTimeSeries(?string $input): array|bool
    {
        $dataout_data = array();
        ErrorOutput::DEBUG("---------------------------");
    
        if ($array = Common::hasJSON($input)) {
            $data = $array["result"][0]["MyEnergyData_MarketDocument"] ?? null;
            $periodstart = $data["period.timeInterval"]["start"] ?? null;
            $periodend = $data["period.timeInterval"]["end"] ?? null;
            $metertimeseries = $data["TimeSeries"][0]["Period"] ?? null;

            ErrorOutput::DEBUG("DATA -> PERIOD START -> " . $periodstart);
            ErrorOutput::DEBUG("DATA -> PERIOD END -> " . $periodend);

            if ($metertimeseries) {
                foreach ($metertimeseries as $day) {
                    $timestart = $day["timeInterval"]["start"];
                    $dt = Common::getDateTime($timestart);

                    if ($day["resolution"] == "PT1H") { //Skip periods with wrong time aggregation
                        foreach ($day["Point"] as $hour) {
                            array_push($dataout_data, array("date" => $dt->format("c"), "kwh" => $hour["out_Quantity.quantity"] )); //Y-m-d\TH:i:s\Z
                            $dt = $dt->add(new \DateInterval("PT1H") ); //Add 1 hour
                        }
                    } else {
                        ErrorOutput::DEBUG("Wrong time period " . $day["resolution"]);
                        $year = Common::getTimeDate(str: $day["timeInterval"]["end"], format: "Y");
                        Configuration::addToIgnore(type: "readings", year: $year);
                    }
                }
            }
            $arrayout["data"] = $dataout_data;
            return $arrayout;
        }
        return false;
    }

    
    /**
     * Tries to fetch data from Eloverblik.dk.
     */
    public static function fetchMeterReadings(string $datefrom, string $dateto): string|bool
    {
        $apiurl = Configuration::getApiUrl();
        try {
            if (($accesstoken = Configuration::getAccessToken()) ) {

            $client = new \GuzzleHttp\Client(["base_uri" => $apiurl]);
            $appendix = "api/meterdata/gettimeseries/";
            $datefrom = $datefrom ?? ErrorOutput::DEBUG("fetchMeterReadings() - datefrom was empty");
            $dateto = $dateto ?? ErrorOutput::DEBUG("fetchMeterReadings() - dateto was empty");

            $timeaggr = "Hour";
            $meterid = Configuration::getSettingsEntry("meter_id") ?? ErrorOutput::handleError(ERRORCODES::APP_INVALID_METERID ?? ERRORCODES::UNKNOWN);

            $queryurl = $appendix . $datefrom . "/" . $dateto . "/" . $timeaggr;
            //Todo: use common method; Todo: Check with regex if queryurl equals the correct format
            ErrorOutput::INFO("Date from " . $datefrom . " to " . $dateto . " (meter: " . $meterid. ")");

            $querybody = json_encode( array("meteringPoints" => array("meteringPoint" => array( $meterid ) )), JSON_PRETTY_PRINT);
            if (empty($querybody)) { return false; } //tmp, remove

            ErrorOutput::DEBUG("Query URL: " . $queryurl);
            ErrorOutput::DEBUG("Query Body: " . $querybody);

            Configuration::setAppEntry("query_date_from", Common::getTimeDate(str: $datefrom) );
            Configuration::setAppEntry("query_date_to", Common::getTimeDate(str: $dateto));
        
            $response = $client->request("POST", $queryurl, [
                "headers" => [
                    "user-agent" => "guzzlehttp/1.0",
                    "content-type"     => "application/json",
                    "accept"     => "application/json",
                    "authorization"      => "Bearer " . $accesstoken
                ],
                "body" => $querybody
            ]);


                if ($response->getStatusCode() == "200") {
                    $resp = (string) $response->getBody();
                        if ($j = Common::hasJSON($resp)) {
                        //Todo: Validate data?
                            if (!empty( $j["result"][0]["MyEnergyData_MarketDocument"])) {
                            return $resp;
                            }
                        } else {
                            ErrorOutput::handleError(ERRORCODES::APP_INVALID_RESPONSE ?? ERRORCODES::UNKNOWN );
                        }
                        return false;
                }
                return false;
            }

        } catch (ClientException $e) { //4xx responses
            $exresp = $e->getResponse();
            if ($exresp->getStatusCode() == "401") { //Unauthorized - Access token probably expired (valid for 24h)
                Parser::fetchToken(); //Request new access token
            } else {
                ErrorOutput::handleError(ERRORCODES::tryFrom($exresp->getStatusCode()) ?? ERRORCODES::UNKNOWN);
            }
        }
        catch (ServerException $e) { //5xx responses
            $exresp = $e->getResponse();
            ErrorOutput::handleError(ERRORCODES::tryFrom($exresp->getStatusCode()) ?? ERRORCODES::UNKNOWN);
        }
        return false;
    }




public static function getInstance()
{
    if (self::$_instance == null) {
        self::$_instance = new Parser();
    }
    return self::$_instance;
}


}

